__author__ = 'Lukas Leufen'
__date__ = '2023-06-01'

import logging
from typing import Dict, Union, List

import requests
from requests.adapters import HTTPAdapter, Retry
# from requests.packages.urllib3.util.retry import Retry

from mlair.helpers import filter_dict_by_value, select_from_dict, data_sources, TimeTracking
import pandas as pd
import xarray as xr

DEFAULT_TIME_DIM = "datetime"
DEFAULT_TARGET_DIM = "variables"
DEFAULT_ITER_DIM = "Stations"
DEFAULT_WINDOW_DIM = "window"


def download_data(file_name: str, meta_file: str, station, statistics_per_var, sampling,
                  store_data_locally=True, data_origin: Dict = None, time_dim=DEFAULT_TIME_DIM,
                  target_dim=DEFAULT_TARGET_DIM, iter_dim=DEFAULT_ITER_DIM, window_dim=DEFAULT_WINDOW_DIM) -> [xr.DataArray, pd.DataFrame]:
    """
    Download data from TOAR database using the JOIN interface or load local era5 data.

    Data is transformed to a xarray dataset. If class attribute store_data_locally is true, data is additionally
    stored locally using given names for file and meta file.

    :param file_name: name of file to save data to (containing full path)
    :param meta_file: name of the meta data file (also containing full path)

    :return: downloaded data and its meta data
    """
    df_era5_local, df_toar, df_ifs_local = None, None, None
    meta_era5_local, meta_toar, meta_ifs_local = None, None, None
    if data_origin is not None:
        era5_local_origin = filter_dict_by_value(data_origin, "era5_local", True)
        era5_local_stats = select_from_dict(statistics_per_var, era5_local_origin.keys())
        ifs_local_origin = filter_dict_by_value(data_origin, "ifs", True)
        ifs_local_stats = select_from_dict(statistics_per_var, ifs_local_origin.keys())
        toar_origin = filter_dict_by_value(data_origin, ["era5_local", "ifs"], False)
        toar_stats = select_from_dict(statistics_per_var, toar_origin.keys())
        assert len(era5_local_origin) + len(toar_origin) + len(ifs_local_origin) == len(data_origin)
        assert len(era5_local_stats) + len(toar_stats) + len(ifs_local_stats) == len(statistics_per_var)
    else:
        era5_local_origin, toar_origin, ifs_local_origin = None, None, None
        era5_local_stats, toar_stats, ifs_local_stats = statistics_per_var, statistics_per_var, statistics_per_var

    # load data
    if era5_local_origin is not None and len(era5_local_stats) > 0:
        # load era5 data
        df_era5_local, meta_era5_local = data_sources.era5.load_era5(
            station_name=station, stat_var=era5_local_stats, sampling=sampling, data_origin=era5_local_origin,
            window_dim=window_dim, time_dim=time_dim)
    if ifs_local_origin is not None and len(ifs_local_stats) > 0:
        # load era5 data
        df_ifs_local, meta_ifs5_local = data_sources.ifs.load_ifs(
            station_name=station, stat_var=ifs_local_stats, sampling=sampling, data_origin=ifs_local_origin,
            lead_time_dim=window_dim, initial_time_dim=time_dim)
    if toar_origin is None or len(toar_stats) > 0:  # TODO return toar data as xarray
        # load combined data from toar-data (v2 & v1)
        df_toar, meta_toar = data_sources.toar_data.download_toar(station=station, toar_stats=toar_stats,
                                                                  sampling=sampling, data_origin=toar_origin)

    valid_df = [e for e in [df_era5_local, df_toar, df_ifs_local] if e is not None]
    if len(valid_df) == 0:
        raise EmptyQueryResult(f"No data available for era5_local, toar-data and ifs_local")
    df = xr.concat(valid_df, dim=time_dim)
    valid_meta = [e for e in [meta_era5_local, meta_toar, meta_ifs_local] if e is not None]
    if len(valid_meta) > 0:
        meta = valid_meta[0]
        for e in valid_meta[1:]:
            meta = meta.combine_first(e)
    else:
        meta = None
    meta.loc["data_origin"] = str(data_origin)
    meta.loc["statistics_per_var"] = str(statistics_per_var)

    xarr = df.expand_dims({iter_dim: station})
    if len(xarr.coords[window_dim]) <= 1:  # keep window dim only if there is more than a single entry
        xarr = xarr.squeeze(window_dim, drop=True)
    if store_data_locally is True:
        # save locally as nc/csv file
        xarr.to_netcdf(path=file_name)
        meta.to_csv(meta_file)
    return xarr, meta


class EmptyQueryResult(Exception):
    """Exception that get raised if a query to JOIN returns empty results."""

    pass


def get_data(opts: Dict, headers: Dict, as_json: bool = True) -> Union[Dict, List, str]:
    """
    Download join data using requests framework.

    Data is returned as json like structure. Depending on the response structure, this can lead to a list or dictionary.

    :param opts: options to create the request url
    :param headers: additional headers information like authorization, can be empty
    :param as_json: extract response as json if true (default True)

    :return: requested data (either as list or dictionary)
    """
    url = create_url(**opts)
    try:
        with TimeTracking(name=url):
            response = retries_session().get(url, headers=headers, timeout=(5, None))  # timeout=(open, read)
        if response.status_code == 200:
            return response.json() if as_json is True else response.text
        else:
            raise EmptyQueryResult(f"There was an error (STATUS {response.status_code}) for request {url}")
    except requests.exceptions.RetryError as e:
        raise EmptyQueryResult(f"There was an RetryError for request {url}: {e}")


def correct_stat_name(stat: str) -> str:
    """
    Map given statistic name to new namespace defined by mapping dict.

    Return given name stat if not element of mapping namespace.

    :param stat: namespace from JOIN server

    :return: stat mapped to local namespace
    """
    mapping = {'average_values': 'mean', 'maximum': 'max', 'minimum': 'min'}
    return mapping.get(stat, stat)


def create_url(base: str, service: str, param_id: Union[str, int, None] = None,
               **kwargs: Union[str, int, float, None]) -> str:
    """
    Create a request url with given base url, service type and arbitrarily many additional keyword arguments.

    :param base: basic url of the rest service
    :param service: service type, e.g. series, stats
    :param param_id: id for a distinct service, is added between ending / of service and ? of kwargs
    :param kwargs: keyword pairs for optional request specifications, e.g. 'statistics=maximum'

    :return: combined url as string
    """
    url = f"{base}"
    if not url.endswith("/"):
        url += "/"
    if service is not None:
        url = f"{url}{service}"
    if not url.endswith("/"):
        url += "/"
    if param_id is not None:
        url = f"{url}{param_id}"
    if len(kwargs) > 0:
        url = f"{url}?{'&'.join(f'{k}={v}' for k, v in kwargs.items() if v is not None)}"
    return url


def retries_session(max_retries=5):
    retry_strategy = Retry(total=max_retries,
                           backoff_factor=1,
                           status_forcelist=[429, 500, 502, 503, 504],
                           method_whitelist=["HEAD", "GET", "OPTIONS"])
    adapter = HTTPAdapter(max_retries=retry_strategy)
    http = requests.Session()
    http.mount("https://", adapter)
    http.mount("http://", adapter)
    return http
